import imagehash
from PIL import Image

from ._hash import _convert_numpy_hash_to_str

"""----------逻辑函数----------"""


def _is_pure_color_image(image_path: str, threshold: float = 0.9) -> bool:
    """是否为纯色图片
    :param image_path: 本地图片路径
    :param threshold: 判断阈值，0~1"""
    try:
        image_pil = Image.open(image_path)
        image_pil = image_pil.convert('L')  # 转灰度图
    except OSError:  # 如果图片损坏，会抛出异常OSError: image file is truncated (4 bytes not processed)
        return False

    dhash = imagehash.average_hash(image_pil, hash_size=16)
    image_pil.close()
    hash_str = _convert_numpy_hash_to_str(dhash)

    proportion = hash_str.count('0') / len(hash_str)
    if proportion > threshold or proportion < (1 - threshold):
        return True
    else:
        return False


"""----------调用函数----------"""


def is_pure_color_image(image_path: str, threshold: float = 0.9) -> bool:
    """是否为纯色图片
    :param image_path: 本地图片路径
    :param threshold: 判断阈值，0~1"""
    return _is_pure_color_image(image_path, threshold)



import io
import os

import numpy
from PIL import Image

"""----------逻辑函数----------"""


def _save_bytes_image(bytes_image: bytes, dirpath: str, filetitle: str) -> str:
    """将一个bytes图片对象保存至本地
    :param bytes_image: bytes图片对象
    :param dirpath: 需要保存到的目录
    :param filetitle: 保存的文件标题（不含文件扩展名）
    :return: 保存到本地的图片路径"""
    image = Image.open(io.BytesIO(bytes_image))

    # 清除元数据
    image.info.clear()

    # 转换图像模式，防止报错OSError: cannot write mode P as JPEG
    image = image.convert('RGB')

    # 提取文件扩展名
    file_extension = image.format  # 文件扩展名带.

    # 如果未提取到文件后缀，则使用jpg格式
    if not file_extension:
        file_extension = '.jpg'

    # 组合保存路径
    filename = filetitle + file_extension
    save_path = os.path.normpath(os.path.join(dirpath, filename))

    # 保存到本地
    if not os.path.exists(dirpath):
        os.mkdir(save_path)
    image.save(save_path)
    image.close()

    return save_path


def _save_numpy_image(numpy_image: numpy.ndarray, dirpath: str, filetitle: str) -> str:
    """将一个numpy图片对象保存至本地
    :param numpy_image: numpy图片对象
    :param dirpath: 需要保存到的目录
    :param filetitle: 保存的文件标题（不含文件扩展名）
    :return: 保存到本地的图片路径"""
    # 转换为uint8类型
    numpy_image = numpy_image.astype(numpy.uint8)

    # 转换为Pillow图像对象
    image = Image.fromarray(numpy_image)

    # 清除元数据
    image.info.clear()

    # 转换图像模式，防止报错OSError: cannot write mode P as JPEG
    image = image.convert('RGB')

    # 提取文件扩展名
    file_extension = image.format  # 文件扩展名带.

    # 如果未提取到文件后缀，则使用jpg格式
    if not file_extension:
        file_extension = '.jpg'

    # 组合保存路径
    filename = filetitle + file_extension
    save_path = os.path.normpath(os.path.join(dirpath, filename))

    # 保存到本地
    if not os.path.exists(dirpath):
        os.mkdir(save_path)
    image.save(save_path)
    image.close()

    return save_path


"""----------调用函数----------"""


def save_bytes_image(bytes_image: bytes, dirpath: str, filetitle: str) -> str:
    """将一个bytes图片对象保存至本地
    :param bytes_image: bytes图片对象
    :param dirpath: 需要保存到的目录
    :param filetitle: 保存的文件标题（不含文件扩展名）
    :return: 保存到本地的图片路径"""
    return _save_bytes_image(bytes_image, dirpath, filetitle)


def save_numpy_image(numpy_image: numpy.ndarray, dirpath: str, filetitle: str) -> str:
    """将一个numpy图片对象保存至本地
    :param numpy_image: numpy图片对象
    :param dirpath: 需要保存到的目录
    :param filetitle: 保存的文件标题（不含文件扩展名）
    :return: 保存到本地的图片路径"""
    return _save_numpy_image(numpy_image, dirpath, filetitle)



import os
from typing import Union

import cv2
import numpy

"""----------逻辑函数----------"""


def _read_local_image_to_numpy(image_path: str) -> Union[numpy.ndarray, None]:
    """读取本地图片，返回numpy对象
    :param image_path: 本地图片路径
    :return: numpy对象"""
    if os.path.exists(image_path):
        image_numpy = cv2.imdecode(numpy.fromfile(image_path, dtype=numpy.uint8), -1)
        return image_numpy
    else:
        return None


def _read_local_image_to_bytes(image_path: str) -> Union[bytes, None]:
    """读取本地图片，返回bytes对象
    :param image_path: 本地图片路径
    :return: bytes对象"""
    if os.path.exists(image_path):
        with open(image_path, 'rb') as file:
            image_bytes = file.read()
            return image_bytes
    else:
        return None


"""----------调用函数----------"""


def read_local_image_to_numpy(image_path: str) -> Union[numpy.ndarray, None]:
    """读取本地图片，返回numpy对象
    :param image_path: 本地图片路径
    :return: numpy对象"""
    return _read_local_image_to_numpy(image_path)


def read_image_to_numpy(image_path: str) -> Union[numpy.ndarray, None]:
    """读取本地图片，返回numpy对象
    :param image_path: 本地图片路径
    :return: numpy对象"""
    return _read_local_image_to_numpy(image_path)


def read_local_image_to_bytes(image_path: str) -> Union[bytes, None]:
    """读取本地图片，返回bytes对象
    :param image_path: 本地图片路径
    :return: bytes对象"""
    return _read_local_image_to_bytes(image_path)


def read_image_to_bytes(image_path: str) -> Union[bytes, None]:
    """读取本地图片，返回bytes对象
    :param image_path: 本地图片路径
    :return: bytes对象"""
    return _read_local_image_to_bytes(image_path)



import numpy

"""----------逻辑函数----------"""


def _calc_ssim(numpy_image_1: numpy.ndarray, numpy_image_2: numpy.ndarray) -> float:
    """计算两个numpy图片对象的SSIM相似值
    :param numpy_image_1: numpy图片对象
    :param numpy_image_2: numpy图片对象
    :return: SSIM相似值
    """
    # 计算均值、方差和协方差
    mean1, mean2 = numpy.mean(numpy_image_1), numpy.mean(numpy_image_2)
    var1, var2 = numpy.var(numpy_image_1), numpy.var(numpy_image_2)
    covar = numpy.cov(numpy_image_1.flatten(), numpy_image_2.flatten())[0][1]

    # 设置常数
    c1 = (0.01 * 255) ** 2
    c2 = (0.03 * 255) ** 2

    # 计算SSIM
    numerator = (2 * mean1 * mean2 + c1) * (2 * covar + c2)
    denominator = (mean1 ** 2 + mean2 ** 2 + c1) * (var1 + var2 + c2)
    ssim = numerator / denominator

    return ssim


"""----------调用函数----------"""


def calc_ssim(numpy_image_1: numpy.ndarray, numpy_image_2: numpy.ndarray) -> float:
    """计算两个numpy图片对象的SSIM相似值
    :param numpy_image_1: numpy图片对象
    :param numpy_image_2: numpy图片对象
    :return: SSIM相似值
    """
    return _calc_ssim(numpy_image_1, numpy_image_2)



from typing import Literal
from typing import Union

import imagehash
from PIL import ImageFile

HASH_TYPE = Literal['ahash', 'phash', 'dhash', 'all']

"""----------逻辑函数----------"""


def _calc_hash(pillow_image: ImageFile, hash_type: HASH_TYPE = 'ahash', hash_size: int = 8) -> dict:
    """计算图片的3种图片Hash值
    :param pillow_image: PIL.ImageFile图片对象
    :param hash_type: 需要计算的hash类型，ahash/phash/dhash/all
    :param hash_size: 计算图片的边长，最终hash值为边长的平方
    :return: {'ahash':None,'phash':None,'dhash':None}
    """
    hash_dict = {'ahash': None, 'phash': None, 'dhash': None}

    if hash_type.lower() == 'all' or hash_type.lower() == 'ahash':
        # 计算均值哈希
        ahash = imagehash.average_hash(pillow_image, hash_size=hash_size)
        ahash_str = _convert_numpy_hash_to_str(ahash)
        hash_dict['ahash'] = ahash_str

    if hash_type.lower() == 'all' or hash_type.lower() == 'phash':
        # 感知哈希
        phash = imagehash.phash(pillow_image, hash_size=hash_size)
        phash_str = _convert_numpy_hash_to_str(phash)
        hash_dict['phash'] = phash_str

    if hash_type.lower() == 'all' or hash_type.lower() == 'dhash':
        # 差异哈希
        dhash = imagehash.dhash(pillow_image, hash_size=hash_size)
        dhash_str = _convert_numpy_hash_to_str(dhash)
        hash_dict['dhash'] = dhash_str

    return hash_dict


def _convert_numpy_hash_to_str(numpy_hash: Union[imagehash.NDArray, imagehash.ImageHash]) -> Union[str, None]:
    """将numpy数组形式的图片Hash值(imagehash.hash)转换为由01组成的字符串
    :param numpy_hash: numpy数组形式的图片Hash值
    :return: 由01组成的字符串"""
    if not numpy_hash:
        return None
    if isinstance(numpy_hash, imagehash.ImageHash):
        numpy_hash = numpy_hash.hash

    hash_str = ''
    for row in numpy_hash:
        for col in row:
            if col:
                hash_str += '1'
            else:
                hash_str += '0'

    return hash_str


def _calc_hamming_distance(hash_1: str, hash_2: str) -> Union[bool, int]:
    """计算两个由01组成的字符串形式的图片Hash值之间的汉明距离
    :param hash_1: 字符串形式的hash值
    :param hash_2: 字符串形式的hash值
    :return: 汉明距离，或者计算失败时返回False
    """
    if len(hash_1) != len(hash_2):  # 位数不同时，无法进行比较
        raise False

    hamming_distance = sum(ch1 != ch2 for ch1, ch2 in zip(hash_1, hash_2))

    return hamming_distance


def _calc_hash_similar(hash_1: str, hash_2: str) -> Union[bool, float]:
    """计算两个由01组成的字符串形式的图片Hash值之间的相似度
    :param hash_1: 字符串形式的hash值
    :param hash_2: 字符串形式的hash值
    :return: 相似度（0~1），或者计算失败时返回False
    """
    if len(hash_1) != len(hash_2):  # 位数不同时，无法进行比较
        raise False

    hash_int1 = int(hash_1, 2)
    hash_int2 = int(hash_2, 2)
    # 使用异或操作计算差异位数
    diff_bits = bin(hash_int1 ^ hash_int2).count('1')
    # 计算相似性
    similarity = 1 - diff_bits / len(hash_1)

    return similarity


"""----------调用函数----------"""


def calc_hash(pillow_image: ImageFile, hash_type: HASH_TYPE = 'ahash', hash_size: int = 8) -> dict:
    """计算图片的3种图片Hash值
    :param pillow_image: PIL.ImageFile图片对象
    :param hash_type: 需要计算的hash类型，ahash/phash/dhash/all
    :param hash_size: 计算图片的边长，最终hash值为边长的平方
    :return: {'ahash':None,'phash':None,'dhash':None}
    """
    return _calc_hash(pillow_image, hash_type, hash_size)


def convert_numpy_hash_to_str(numpy_hash: Union[imagehash.NDArray, imagehash.ImageHash]) -> Union[str, None]:
    """将numpy数组形式的图片Hash值(imagehash.hash)转换为由01组成的字符串
    :param numpy_hash: numpy数组形式的图片Hash值
    :return: 由01组成的字符串"""
    return _convert_numpy_hash_to_str(numpy_hash)


def numpy_hash_to_str(numpy_hash: Union[imagehash.NDArray, imagehash.ImageHash]) -> Union[str, None]:
    """将numpy数组形式的图片Hash值(imagehash.hash)转换为由01组成的字符串
    :param numpy_hash: numpy数组形式的图片Hash值
    :return: 由01组成的字符串"""
    return _convert_numpy_hash_to_str(numpy_hash)


def calc_hamming_distance(hash_1: str, hash_2: str) -> Union[bool, int]:
    """计算两个由01组成的字符串形式的图片Hash值之间的汉明距离
    :param hash_1: 字符串形式的hash值
    :param hash_2: 字符串形式的hash值
    :return: 汉明距离，或者计算失败时返回False
    """
    return _calc_hamming_distance(hash_1, hash_2)


def calc_hash_hamming_distance(hash_1: str, hash_2: str) -> Union[bool, int]:
    """计算两个由01组成的字符串形式的图片Hash值之间的汉明距离
    :param hash_1: 字符串形式的hash值
    :param hash_2: 字符串形式的hash值
    :return: 汉明距离，或者计算失败时返回False
    """
    return _calc_hamming_distance(hash_1, hash_2)


def calc_hash_similar(hash_1: str, hash_2: str) -> Union[bool, float]:
    """计算两个由01组成的字符串形式的图片Hash值之间的相似度
    :param hash_1: 字符串形式的hash值
    :param hash_2: 字符串形式的hash值
    :return: 相似度（0~1），或者计算失败时返回False
    """
    return _calc_hash_similar(hash_1, hash_2)
